#include <chrono>
#include <iomanip>
#include <random>
#include <memory>
#include <vector>
#include <iostream>
#include <algorithm>
#include <loki/SmallObj.h>

using namespace std;


template<unsigned int N>
struct MyData
#ifdef USE_LOKI_SOA
    : public Loki::SmallValueObject<>
#else
#ifdef SYSTEM_ALLOC
    // do not use any specific allocator
#else
#error USE_LOKI_SOA or NOTHING must be defined
#endif
#endif
{
  static_assert(N<=LOKI_MAX_SMALL_OBJECT_SIZE, "N must be less than or equal to LOKI_MAX_SMALL_OBJECT_SIZE");

  MyData(void)
  {
    data_[N-1]=0;
  }

  char data_[N];
};


typedef chrono::high_resolution_clock Clock;

double diff(const Clock::time_point begin)
{
  const Clock::time_point end=Clock::now();
  const double diff=chrono::microseconds(end-begin).count()/(1000.0*1000);
  return diff;
}


int main(void)
{
  typedef MyData<DATA_SIZE> Data;
  typedef unique_ptr<Data>  DataPtr;
  typedef vector<DataPtr>   Vec;

  constexpr Vec::size_type size=1*1000*1000;
  Vec v;

  // fill up
  const auto t1=Clock::now();
  v.reserve(size);
  for(Vec::size_type i=0; i<size; ++i)
    v.push_back( DataPtr{new Data{}} );
  const double e1=diff(t1);

  // shuffle
  const auto t2=Clock::now();
  if(true)
  {
    random_device rd;
    //mt19937 enginei(rdi());
    knuth_b eng( rd() );
    shuffle( v.begin(), v.end(), eng );
  }
  const double e2=diff(t2);

  int out=0;    // prevent from compiling out stuff
  // erase
  const auto t3=Clock::now();
  for(Vec::size_type i=0; i<size; ++i)
  {
    auto lastIt=v.begin()+v.size()-1;
    out+=(*lastIt)->data_[0];
    v.erase(lastIt);
  }
  const double e3=diff(t3);

  cout << fixed << setprecision(3) << e1 << " | "
       << fixed << setprecision(3) << e2 << " | "
       << fixed << setprecision(3) << e3 << " | "
       << endl;

  return out%42;
}
